/*

mfloppy.c - Memotech Floppy tool

When experimenting with fdutils, I found this best for looking at type 07s.
setfdprm /dev/fd0 sect=16 hd ssize=256 cyl=80 dtr=2 tpi=96

*/

/*...sincludes:0:*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <linux/fd.h>
#include <linux/fdreg.h>
#include <linux/major.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
/*...e*/

/*...sfatal:0:*/
static void fatal(const char *fmt, ...)
	{
	va_list	vars;
	char s[256+1];
	va_start(vars, fmt);
	vsprintf(s, fmt, vars);
	va_end(vars);
	fprintf(stderr, "mfloppy: %s\n", s);
	exit(1);
	}
/*...e*/
/*...susage:0:*/
static void usage(void)
	{
	fprintf(stderr, "usage: mfloppy read|verify|write|write+verify /dev/fdX file.mfloppy\n");
	exit(1);
	}
/*...e*/

#define	RATE 2

/*...sraw_read:0:*/
static int raw_read(
	int fd,
	int ptrack,
	int ltrack,
	int head,
	int sector,
	unsigned char *data
	)
	{
	int tries;
	for ( tries = 1; tries < 100; tries++ )
		{
		struct floppy_raw_cmd r;
		int rc;
		r.rate = RATE;
		r.flags = FD_RAW_READ|FD_RAW_INTR|FD_RAW_NEED_DISK|FD_RAW_NEED_SEEK;
		r.length = 256;
		r.data = data;
		r.track = ptrack;
		r.cmd_count = 9;
		r.cmd[0] = FD_READ;
		r.cmd[1] = head * 4; /* 0 or 4 */
		r.cmd[2] = ltrack; /* cylinder */
		r.cmd[3] = head; /* 0 or 1 */
		r.cmd[4] = sector;
		r.cmd[5] = 1; /* sector size = 256 bytes */
		r.cmd[6] = 16; /* sectors per track */
		r.cmd[7] = 0x1b; /* "just in case" */
		r.cmd[8] = 0xff; /* "usually" */
		memset(data, 0xdd, 256);
		if ( (rc = ioctl(fd, FDRAWCMD, &r)) < 0 )
			return rc;
		if ( r.length == 0 )
			{
			if ( tries > 10 )
				fprintf(stderr, "track %d head %d sector %d read after %d tries\n", ltrack, head, sector, tries); 
			return 0;
			}
		}
	fprintf(stderr, "track %d head %d sector %d could not be read after %d tries\n", ltrack, head, sector, tries);
	return 0;
	}
/*...e*/
/*...sraw_verify:0:*/
static int raw_verify(
	int fd,
	int ptrack,
	int ltrack,
	int head,
	int sector,
	unsigned char *data
	)
	{
	int tries;
	for ( tries = 1; tries < 20; tries++ )
		{
		struct floppy_raw_cmd r;
		unsigned char data2[256];
		int rc;
		r.rate = RATE;
		r.flags = FD_RAW_READ|FD_RAW_INTR|FD_RAW_NEED_DISK|FD_RAW_NEED_SEEK;
		r.length = 256;
		r.data = data2;
		r.track = ptrack;
		r.cmd_count = 9;
		r.cmd[0] = FD_READ;
		r.cmd[1] = head * 4; /* 0 or 4 */
		r.cmd[2] = ltrack; /* cylinder */
		r.cmd[3] = head; /* 0 or 1 */
		r.cmd[4] = sector;
		r.cmd[5] = 1; /* sector size = 256 bytes */
		r.cmd[6] = 16; /* sectors per track */
		r.cmd[7] = 0x1b; /* "just in case" */
		r.cmd[8] = 0xff; /* "usually" */
		memset(data2, 0xdd, 256);
		if ( (rc = ioctl(fd, FDRAWCMD, &r)) < 0 )
			return rc;
		if ( r.length == 0 )
			{
			if ( !memcmp(data, data2, 256) )
				{
				if ( tries > 10 )
					fprintf(stderr, "track %d head %d sector %d verified after %d tries\n", ltrack, head, sector, tries); 
				return 0;
				}
			}
		}
	fprintf(stderr, "track %d head %d sector %d could not be verified after %d tries\n", ltrack, head, sector, tries);
	return 0;
	}
/*...e*/
/*...sraw_write:0:*/
static int raw_write(
	int fd,
	int ptrack,
	int ltrack,
	int head,
	int sector,
	unsigned char *data
	)
	{
	int tries;
	for ( tries = 1; tries < 100; tries++ )
		{
		struct floppy_raw_cmd r;
		int rc;
		r.rate = RATE;
		r.flags = FD_RAW_WRITE|FD_RAW_INTR|FD_RAW_NEED_DISK|FD_RAW_NEED_SEEK;
		r.length = 256;
		r.data = data;
		r.track = ptrack;
		r.cmd_count = 9;
		r.cmd[0] = FD_WRITE;
		r.cmd[1] = head * 4; /* 0 or 4 */
		r.cmd[2] = ltrack; /* cylinder */
		r.cmd[3] = head; /* 0 or 1 */
		r.cmd[4] = sector;
		r.cmd[5] = 1; /* sector size = 256 bytes */
		r.cmd[6] = 16; /* sectors per track */
		r.cmd[7] = 0x1b; /* "just in case" */
		r.cmd[8] = 0xff; /* "usually" */
		if ( (rc = ioctl(fd, FDRAWCMD, &r)) < 0 )
			return rc;
		if ( r.length == 0 )
			{
			if ( tries > 10 )
				fprintf(stderr, "track %d head %d sector %d written after %d tries\n", ltrack, head, sector, tries); 
			return 0;
			}
		}
	fprintf(stderr, "track %d head %d sector %d could not be written after %d tries\n", ltrack, head, sector, tries);
	return 0;
	}
/*...e*/
/*...sraw_write_and_verify:0:*/
static int raw_write_and_verify(
	int fd,
	int ptrack,
	int ltrack,
	int head,
	int sector,
	unsigned char *data
	)
	{
	int attempts;
	for ( attempts = 0; attempts < 20; attempts++ )
		{
		unsigned char data2[256];
		int rc;
		if ( (rc = raw_write(fd, ptrack, ltrack, head, sector, data)) < 0 )
			return rc;
		if ( (rc = raw_read(fd, ptrack, ltrack, head, sector, data2)) < 0 )
			return rc;
		if ( !memcmp(data, data2, 256) )
			{
			if ( attempts > 3 )
				fprintf(stderr, "track %d head %d sector %d written and verified after %d attempts\n", ltrack, head, sector, attempts);
			return 0;
			}
		}
	fprintf(stderr, "track %d head %d sector %d write and verify failed after %d attempts\n", ltrack, head, sector, attempts);
	return 0;
	}
/*...e*/

/*...smain:0:*/
int main(int argc, char *argv[])
	{
	int fd, n_tracks, step, rc;
	struct floppy_raw_cmd r;
	struct stat buf;
	int drive;

	if ( argc != 4 )
		usage();

	if ( strcmp(argv[1], "read"        ) &&
	     strcmp(argv[1], "verify"      ) &&
	     strcmp(argv[1], "write"       ) &&
	     strcmp(argv[1], "write+verify") &&
	     strcmp(argv[1], "format"      ) )
		usage();

	if ( (fd = open(argv[2], O_ACCMODE|O_NDELAY)) == -1 )
		fatal("can't open %s", argv[2]);

	if ( fstat(fd, &buf) < 0 )
		{
		close(fd);
		fatal("can't fstat %s", argv[2]);
		}

	if ( major(buf.st_rdev) != FLOPPY_MAJOR )
		{
		close(fd);
		fatal("%s is not a floppy", argv[2]);
		}

	drive = minor(buf.st_rdev);
	drive = (drive & 3) + ((drive & 0x80) >> 5);

	if ( (rc = ioctl(fd, FDRESET, NULL)) < 0 )
		{
		close(fd);
		fatal("can't reset floppy: rc=%d", rc);
		}

	r.rate = RATE;
	r.flags = FD_RAW_INTR;
	r.cmd_count = 2;
	r.cmd[0] = FD_RECALIBRATE;
	r.cmd[1] = drive;
	if ( (rc = ioctl(fd, FDRAWCMD, &r)) < 0 )
		{
		close(fd);
		fatal("can't recalibrate floppy: rc=%d", rc);
		}

	/* Check that it is D/S D/D */
	r.rate = RATE;
	r.flags = FD_RAW_INTR|FD_RAW_NEED_DISK|FD_RAW_NEED_SEEK;
	r.length = 0;
	r.data = NULL;
	r.track = 0;
	r.cmd_count = 2;
	r.cmd[0] = FD_READID;
	r.cmd[1] = 4; /* 0 or 4 */
	if ( (rc = ioctl(fd, FDRAWCMD, &r)) < 0 )
		{
		close(fd);
		fatal("can't readid floppy: rc=%d", rc);
		}
	if ( r.reply_count < 7 )
		{
		close(fd);
		fatal("can't readid floppy, truncated reply");
		}
	if ( r.reply[0] > 7 )
		{
		close(fd);
		fatal("floppy is not D/S");
		}
	if ( r.reply[6] != 1 )
		{
		close(fd);
		fatal("floppy is not D/D");
		}

	/* Look to see if 40 (aka S/T) or 80 (aka D/T) track */
	r.rate = RATE;
	r.flags = FD_RAW_INTR|FD_RAW_NEED_DISK|FD_RAW_NEED_SEEK;
	r.length = 0;
	r.data = NULL;
	r.track = 2;
	r.cmd_count = 2;
	r.cmd[0] = FD_READID;
	r.cmd[1] = 0; /* 0 or 4 */
	if ( (rc = ioctl(fd, FDRAWCMD, &r)) < 0 )
		{
		close(fd);
		fatal("can't readid floppy: rc=%d", rc);
		}
	if ( r.reply_count < 7 )
		{
		close(fd);
		fatal("can't readid floppy, truncated reply");
		}
	switch ( r.reply[3] )
		{
		case 2:
			n_tracks = 80;
			step = 1;
			printf("type 07 (D/S, D/D, 80 track, 640KB) floppy detected\n");
			break;
		case 1:
			n_tracks = 40;
			step = 2;
			printf("type 03 (D/S, D/D, 40 track, 320KB) floppy detected\n");
			break;
		default:
			close(fd);
			fatal("floppy isn't 80 or 40 track");
		}

	if ( !strcmp(argv[1], "read") )
/*...sread floppy to file:16:*/
{
int fd2, track, head, sector;
if ( (fd2 = open(argv[3], O_CREAT|O_TRUNC|O_WRONLY, 0644)) == -1 )
	{
	close(fd);
	fatal("can't create %s", argv[3]);
	}
for ( track = 0; track < n_tracks; track++ )
	for ( head = 0; head < 2; head++ )
		for ( sector = 1; sector <= 16; sector++ )
			{
			unsigned char data[256];
			int rc;
			if ( (rc = raw_read(fd, track*step, track, head, sector, data)) < 0 )
				{
				close(fd2);
				close(fd);
				fatal("can't read track %d head %d sector %d, rc=%d", track, head, sector, rc);
				}
			if ( write(fd2, data, 256) != 256 )
				{
				close(fd2);
				close(fd);
				fatal("error writing file");
				}
			}
close(fd2);
}
/*...e*/
	else if ( !strcmp(argv[1], "verify") )
/*...sverify floppy against file:16:*/
{
int fd2, track, head, sector;
if ( (fd2 = open(argv[3], O_RDONLY)) == -1 )
	{
	close(fd);
	fatal("can't open %s", argv[3]);
	}
for ( track = 0; track < n_tracks; track++ )
	for ( head = 0; head < 2; head++ )
		for ( sector = 1; sector <= 16; sector++ )
			{
			unsigned char data[256];
			int rc;
			if ( read(fd2, data, 256) != 256 )
				{
				close(fd2);
				close(fd);
				fatal("error reading file");
				}
			if ( (rc = raw_verify(fd, track*step, track, head, sector, data)) < 0 )
				{
				close(fd2);
				close(fd);
				fatal("can't verify track %d head %d sector %d, rc=%d", track, head, sector, rc);
				}
			}
close(fd2);
}
/*...e*/
	else if ( !strcmp(argv[1], "write") )
/*...swrite file to floppy:16:*/
{
int fd2, track, head, sector;
if ( (fd2 = open(argv[3], O_RDONLY)) == -1 )
	{
	close(fd);
	fatal("can't open %s", argv[3]);
	}
for ( track = 0; track < n_tracks; track++ )
	for ( head = 0; head < 2; head++ )
		for ( sector = 1; sector <= 16; sector++ )
			{
			unsigned char data[256];
			if ( read(fd2, data, 256) != 256 )
				{
				close(fd2);
				close(fd);
				fatal("error reading file");
				}
			if ( (rc = raw_write(fd, track*step, track, head, sector, data)) < 0 )
				{
				close(fd2);
				close(fd);
				fatal("can't write track %d head %d sector %d, rc=%d", track, head, sector, rc);
				}
			}
close(fd2);
}
/*...e*/
	else if ( !strcmp(argv[1], "write+verify") )
/*...swrite file to floppy and verify:16:*/
{
int fd2, track, head, sector;
if ( (fd2 = open(argv[3], O_RDONLY)) == -1 )
	{
	close(fd);
	fatal("can't open %s", argv[3]);
	}
for ( track = 0; track < n_tracks; track++ )
	for ( head = 0; head < 2; head++ )
		for ( sector = 1; sector <= 16; sector++ )
			{
			unsigned char data[256];
			if ( read(fd2, data, 256) != 256 )
				{
				close(fd2);
				close(fd);
				fatal("error reading file");
				}
			if ( (rc = raw_write_and_verify(fd, track*step, track, head, sector, data)) < 0 )
				{
				close(fd2);
				close(fd);
				fatal("can't write and verify track %d head %d sector %d, rc=%d", track, head, sector, rc);
				}
			}
close(fd2);
}
/*...e*/
	else
/*...sformat \64\\64\\64\ attempt 1 \45\ not working:16:*/
/* This doesn't work, which is why the format command is undocumented.
   So at the moment I recommend the use of superformat from fdutils.
   I might come back and revisit this at some point. */

{
struct floppy_struct param;
int rc, track, head;

param.gap = 0;
param.spec1 = 0;
param.fmt_gap = 0;
param.name = NULL;

if ( (rc = ioctl(fd, FDGETPRM, &param)) < 0 )
	{
	close(fd);
	fatal("can't get floppy parameters: rc=%d", rc);
	}

printf("track=%02x head=%02x sect=%02x rate=%02x fmt_gap=%02x\n",
	param.track, param.head, param.sect, param.rate, param.fmt_gap);

param.track = 80;
param.head = 2;
param.sect = 16;
param.size = param.track * param.head * param.sect;
param.stretch = 0;
param.rate = (((1-2)<<3)&FD_SIZECODEMASK) | RATE;
	/* This is a really dodgy
	   sizecodes intended to be passed are 2, 3, 4, 5, 6, 7 */

if ( (rc = ioctl(fd, FDSETPRM, &param)) < 0 )
	{
	close(fd);
	fatal("can't set floppy parameters: rc=%d", rc);
	}

if ( (rc = ioctl(fd, FDFMTBEG, NULL)) < 0 )
	{
	close(fd);
	fatal("can't begin formatting: rc=%d", rc);
	}

for ( track = 0; track < n_tracks; track++ )
	for ( head = 0; head < 2; head++ )
		{
		struct format_descr descr;
		descr.device = 0; /* @@@ should this be drive? */
		descr.track = track*step;
		descr.head  = head;
		if ( (rc = ioctl(fd, FDFMTTRK, &descr)) < 0 )
			{
			ioctl(fd, FDFMTEND, NULL);
			close(fd);
			fatal("error formatting track %d head %d: rc=%d", track, head, rc);
			}
		}

if ( (rc = ioctl(fd, FDFMTEND, NULL)) < 0 )
	{
	close(fd);
	fatal("can't end formatting: rc=%d", rc);
	}
}
/*...e*/
	close(fd);
	return 0;
	}
/*...e*/
